from __future__ import annotations

import sys
from functools import wraps
from typing import Any, Callable, Iterable, Union
from pathlib import Path
import json

from pyflowlauncher.shared import logger

from .event import EventHandler
from .jsonrpc import JsonRPCClient
from .result import JsonRPCAction, ResultResponse
from .manifest import PluginManifestSchema, MANIFEST_FILE

Method = Callable[..., Union[ResultResponse, JsonRPCAction]]


class Plugin:

    def __init__(self, methods: list[Method] | None = None) -> None:
        self._logger = logger(self)
        self._client = JsonRPCClient()
        self._event_handler = EventHandler()
        self._settings: dict[str, Any] = {}
        if methods:
            self.add_methods(methods)

    def add_method(self, method: Method) -> None:
        """Add a method to the event handler."""
        name = getattr(method, '__name__', method.__class__.__name__).lower()
        self._logger.debug(f"Adding method: '{name}'")
        self._event_handler.add_method(method, name=name)

    def add_methods(self, methods: Iterable[Method]) -> None:
        self._event_handler.add_methods(methods)

    def on_method(self, method: Method) -> Method:
        @wraps(method)
        def wrapper(*args, **kwargs):
            return method(*args, **kwargs)
        self._event_handler.add_method(wrapper)
        return wrapper

    @property
    def settings(self) -> dict:
        if self._settings is None:
            self._settings = {}
        self._settings = self._client.recieve().get('settings', {})
        return self._settings

    def run(self) -> None:
        request = self._client.recieve()
        method = request.get('method')
        parameters = request.get('parameters', [])
        feedback = self._event_handler(method, *parameters)
        # Inject settings if changed
        if 'result' in feedback and self._settings is not None:
            feedback['SettingsChange'] = self.settings
        self._client.send(feedback)

    @property
    def run_dir(self) -> Path:
        """Return the run directory of the plugin."""
        return Path(sys.argv[0]).parent

    def root_dir(self) -> Path:
        """Return the root directory of the plugin."""
        current_dir = self.run_dir
        for part in current_dir.parts:
            if current_dir.joinpath(MANIFEST_FILE).exists():
                return current_dir
            current_dir = current_dir.parent
        raise FileNotFoundError(f"Could not find {MANIFEST_FILE} in {self.run_dir} or any parent directory.")

    def manifest(self) -> PluginManifestSchema:
        """Return the plugin manifest."""
        with open(self.root_dir() / MANIFEST_FILE, 'r', encoding='utf-8') as f:
            manifest = json.load(f)
        return manifest



from __future__ import annotations

import json
import sys
from typing import Any, Mapping

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, TypedDict
else:
    from typing import NotRequired, TypedDict


class JsonRPCRequest(TypedDict):
    method: str
    parameters: list
    settings: NotRequired[dict[Any, Any]]


class JsonRPCClient:

    def send(self, data: Mapping) -> None:
        json.dump(data, sys.stdout)

    def recieve(self) -> JsonRPCRequest:
        try:
            return json.loads(sys.argv[1])
        except (IndexError, json.JSONDecodeError):
            return {'method': 'query', 'parameters': ['']}



import importlib.metadata
import logging
import os

from .plugin import Plugin
from .result import JsonRPCAction, Result, send_results, ResultResponse
from .method import Method

log_level = os.environ.get("FLOW_LAUNCHER_API_LOG_LEVEL", "INFO")

logger = logging.getLogger(__name__)


__version__ = importlib.metadata.version(__name__)


__all__ = [
    "Plugin",
    "ResultResponse",
    "send_results",
    "Result",
    "JsonRPCAction",
    "Method",
]


logging.basicConfig(
    level=log_level,
    format="%(asctime)s <%(name)s>[%(levelname)s]: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)





class EventHandler:

    def __init__(self):
        self._methods = {}

    def add_method(self, method, *, name=None):
        self._methods[name or method.__name__] = method

    def add_methods(self, methods):
        for method in methods:
            self.add_method(method)

    def __call__(self, method, *args, **kwargs):
        return self._methods[method](*args, **kwargs)



from __future__ import annotations

import sys
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Union

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, TypedDict
else:
    from typing import NotRequired, TypedDict


if TYPE_CHECKING:
    from .plugin import Method


class JsonRPCAction(TypedDict):
    """Flow Launcher JsonRPCAction"""
    method: str
    parameters: Iterable
    dontHideAfterAction: NotRequired[bool]


class Glyph(TypedDict):
    """Flow Launcher Glyph"""
    Glyph: str
    FontFamily: str


@dataclass
class Result:
    Title: str
    SubTitle: Optional[str] = None
    IcoPath: Optional[Union[str, Path]] = None
    Score: int = 0
    JsonRPCAction: Optional[JsonRPCAction] = None
    ContextData: Optional[Iterable] = None
    Glyph: Optional[Glyph] = None
    CopyText: Optional[str] = None
    AutoCompleteText: Optional[str] = None
    RoundedIcon: bool = False

    def as_dict(self) -> Dict[str, Any]:
        return self.__dict__

    def add_action(self, method: Method,
                   parameters: Optional[Iterable[Any]] = None,
                   *,
                   dont_hide_after_action: bool = False) -> None:
        self.JsonRPCAction = {
            "method": method.__name__,
            "parameters": parameters or [],
            "dontHideAfterAction": dont_hide_after_action
        }


class ResultResponse(TypedDict):
    result: List[Dict[str, Any]]
    SettingsChange: NotRequired[Optional[Dict[str, Any]]]


def send_results(results: Iterable[Result], settings: Optional[Dict[str, Any]] = None) -> ResultResponse:
    """Formats and returns results as a JsonRPCResponse"""
    return {'result': [result.as_dict() for result in results], 'SettingsChange': settings}



from dataclasses import dataclass, field
from typing import List

SPACE_CHAR: str = ' '
QUERY_SEARCH_PRECISION = {
    'Regular': 50,
    'Low': 20,
    'None': 0
}
DEFAULT_QUERY_SEARCH_PRECISION = QUERY_SEARCH_PRECISION['Regular']

"""
This is a python copy of Flow Launcher's string matcher.
I take no credit for the algorithm, I just translated it to python.
"""


@dataclass
class MatchData:
    """Match data"""
    matched: bool
    score_cutoff: int
    index_list: List[int] = field(default_factory=list)
    score: int = 0


def string_matcher(query: str, text: str, ignore_case: bool = True,
                   query_search_precision: int = DEFAULT_QUERY_SEARCH_PRECISION) -> MatchData:
    """Compare query to text"""
    if not text or not query:
        return MatchData(False, query_search_precision)

    query = query.strip()

    current_acronym_query_index = 0
    acronym_match_data: List[int] = []
    acronyms_total_count: int = 0
    acronyms_matched: int = 0

    full_text_lower: str = text.lower() if ignore_case else text
    query_lower: str = query.lower() if ignore_case else query

    query_substrings: List[str] = query_lower.split(' ')
    current_query_substring_index: int = 0
    current_query_substring = query_substrings[current_query_substring_index]
    current_query_substring_char_index = 0

    first_match_index = -1
    first_match_index_in_word = -1
    last_match_index = 0
    all_query_substrings_matched: bool = False
    match_found_in_previous_loop: bool = False
    all_substrings_contained_in_text: bool = True

    index_list: List[int] = []
    space_indices: List[int] = []
    for text_index in range(len(full_text_lower)):
        if (current_acronym_query_index >= len(query_lower) or
                (current_acronym_query_index >= len(query_lower) and all_query_substrings_matched)):
            break

        if full_text_lower[text_index] == SPACE_CHAR and current_query_substring_char_index == 0:
            space_indices.append(text_index)

        if is_acronym(text, text_index):
            if full_text_lower[text_index] == query_lower[current_acronym_query_index]:
                acronym_match_data.append(text_index)
                acronyms_matched += 1
                current_acronym_query_index += 1

        if is_acronym_count(text, text_index):
            acronyms_total_count += 1

        if all_query_substrings_matched or (full_text_lower[text_index]
                                            != current_query_substring[current_query_substring_char_index]):
            match_found_in_previous_loop = False
            continue

        if first_match_index < 0:
            first_match_index = text_index

        if current_query_substring_char_index == 0:
            match_found_in_previous_loop = True
            first_match_index_in_word = text_index
        elif not match_found_in_previous_loop:
            start_index_to_verify = text_index - current_query_substring_char_index

            if all_previous_chars_matched(start_index_to_verify, current_query_substring_char_index, full_text_lower,
                                          current_query_substring):
                match_found_in_previous_loop = True
                first_match_index_in_word = start_index_to_verify if current_query_substring_index == 0 else first_match_index

                index_list = get_updated_index_list(
                    start_index_to_verify, current_query_substring_char_index, first_match_index_in_word, index_list)

        last_match_index = text_index + 1
        index_list.append(text_index)

        current_query_substring_char_index += 1

        if current_query_substring_char_index == len(current_query_substring):
            all_substrings_contained_in_text = match_found_in_previous_loop and all_substrings_contained_in_text

            current_query_substring_index += 1

            all_query_substrings_matched = all_query_substrings_matched_func(
                current_query_substring_index, len(query_substrings))

            if all_query_substrings_matched:
                continue

            current_query_substring = query_substrings[current_query_substring_index]
            current_query_substring_char_index = 0

    if acronyms_matched > 0 and acronyms_matched == len(query):
        acronyms_score: int = int(acronyms_matched * 100 / acronyms_total_count)

        if acronyms_score >= query_search_precision:
            return MatchData(True, query_search_precision, acronym_match_data, acronyms_score)

    if all_query_substrings_matched:

        nearest_space_index = calculate_closest_space_index(
            space_indices, first_match_index)

        score: float = calculate_search_score(query, text, first_match_index - nearest_space_index - 1,
                                              space_indices, last_match_index - first_match_index,
                                              all_substrings_contained_in_text)

        return MatchData(True, query_search_precision, index_list, int(score))

    return MatchData(False, query_search_precision)


def calculate_search_score(query: str, text: str, first_index: int, space_indices: List[int],
                           match_length: int, all_substrings_contained_in_text: bool):
    score = 100 * (len(query) + 1) / ((1 + first_index) + (match_length + 1))

    if first_index == 0 and all_substrings_contained_in_text:
        score -= len(space_indices)

    if (len(text) - len(query)) < 5:
        score += 20
    elif (len(text) - len(query)) < 10:
        score += 10

    if all_substrings_contained_in_text:
        count: int = len(query.replace(' ', ''))
        threshold: int = 4
        if count <= threshold:
            score += count * 10
        else:
            score += threshold * 10 + (count - threshold) * 5

    return score


def get_updated_index_list(start_index_to_verify: int,
                           current_query_substring_char_index: int,
                           first_matched_index_in_word: int, index_list: List[int]):
    updated_list: List[int] = []

    for idx, item in enumerate(index_list):
        if item >= first_matched_index_in_word:
            index_list.pop(idx)

    updated_list.extend(index_list)

    for i in range(current_query_substring_char_index):
        updated_list.append(start_index_to_verify + i)

    return updated_list


def all_query_substrings_matched_func(current_query_substring_index: int, query_substrings_length: int) -> bool:
    return current_query_substring_index >= query_substrings_length


def all_previous_chars_matched(start_index_to_verify: int,
                               current_query_substring_char_index: int,
                               full_text_lower: str, current_query_substring: str) -> bool:
    all_match = True
    for i in range(current_query_substring_char_index):
        if full_text_lower[start_index_to_verify + i] != current_query_substring[i]:
            all_match = False

    return all_match


def is_acronym(text: str, text_index: int) -> bool:
    if is_acronym_char(text, text_index) or is_acronym_number(text, text_index):
        return True
    return False


def is_acronym_count(text: str, text_index: int) -> bool:
    if is_acronym_char(text, text_index):
        return True
    if is_acronym_number(text, text_index):
        return text_index == 0 or text[text_index - 1] == SPACE_CHAR

    return False


def is_acronym_char(text: str, text_index: int) -> bool:
    return text[text_index].isupper() or text_index == 0 or text[text_index - 1] == SPACE_CHAR


def is_acronym_number(text: str, text_index: int) -> bool:
    return text[text_index].isdigit()


def calculate_closest_space_index(space_indices: List[int], first_match_index: int) -> int:

    closest_space_index = -1

    for i in space_indices:
        if i < first_match_index:
            closest_space_index = i
        else:
            break

    return closest_space_index



import logging
import os
from pathlib import Path
from typing import Dict, Optional

_logger = logging.getLogger(__name__)

IMAGE_DIR = "Images"
FLOW_PROGRAM_DIRECTORY = os.getenv("FLOW_PROGRAM_DIRECTORY", None)

ENV_EXISTS: bool = FLOW_PROGRAM_DIRECTORY is not None

if not ENV_EXISTS:
    _logger.warning("Unable to find FLOW_PROGRAM_DIRECTORY environment variable. Icons will not be loaded.")


def _get_icon(icon_name: str, file_ext: str = "png") -> Optional[str]:
    if ENV_EXISTS:
        return str(Path(FLOW_PROGRAM_DIRECTORY) / IMAGE_DIR / f"{icon_name}.{file_ext}")  # type: ignore
    return None


ICONS: Dict[str, Optional[str]] = {
    "2566": _get_icon("2566"),
    "admin": _get_icon("admin"),
    "app": _get_icon("app"),
    "app_error": _get_icon("app_error"),
    "app_missing_img": _get_icon("app_missing_img"),
    "baidu": _get_icon("baidu"),
    "bing": _get_icon("bing"),
    "bookmark": _get_icon("bookmark"),
    "browser": _get_icon("Browser"),
    "calculator": _get_icon("calculator"),
    "cancel": _get_icon("cancel"),
    "checkupdate": _get_icon("checkupdate"),
    "close": _get_icon("close"),
    "cmd": _get_icon("cmd"),
    "color": _get_icon("color"),
    "context_menu": _get_icon("context_menu"),
    "controlpanel_small": _get_icon("ControlPanel_Small"),
    "copy": _get_icon("copy"),
    "copylink": _get_icon("copylink"),
    "deletedfolder": _get_icon("deletedfolder"),
    "disable": _get_icon("disable"),
    "down": _get_icon("down"),
    "duckduckgo": _get_icon("duckduckgo"),
    "error": _get_icon("error"),
    "everything_error": _get_icon("everything_error"),
    "excludeindexpath": _get_icon("excludeindexpath"),
    "exe": _get_icon("EXE"),
    "explorer": _get_icon("explorer"),
    "facebook": _get_icon("facebook"),
    "file": _get_icon("file"),
    "find": _get_icon("find"),
    "folder": _get_icon("folder"),
    "gamemode": _get_icon("gamemode"),
    "gist": _get_icon("gist"),
    "github": _get_icon("github"),
    "gmail": _get_icon("gmail"),
    "google": _get_icon("google"),
    "google_drive": _get_icon("google_drive"),
    "google_maps": _get_icon("google_maps"),
    "google_translate": _get_icon("google_translate"),
    "hibernate": _get_icon("hibernate"),
    "history": _get_icon("history"),
    "image": _get_icon("image"),
    "index_error": _get_icon("index_error"),
    "index_error2": _get_icon("index_error2"),
    "indexoption": _get_icon("indexoption"),
    "link": _get_icon("link"),
    "loading": _get_icon("loading"),
    "lock": _get_icon("lock"),
    "logoff": _get_icon("logoff"),
    "manifestsite": _get_icon("manifestsite"),
    "netflix": _get_icon("netflix"),
    "new_message": _get_icon("New Message"),
    "ok": _get_icon("ok"),
    "open": _get_icon("open"),
    "openrecyclebin": _get_icon("openrecyclebin"),
    "pictures": _get_icon("pictures"),
    "pluginsmanager": _get_icon("pluginsmanager"),
    "program": _get_icon("program"),
    "quickaccess": _get_icon("quickaccess"),
    "recyclebin": _get_icon("recyclebin"),
    "removequickaccess": _get_icon("removequickaccess"),
    "request": _get_icon("request"),
    "restart": _get_icon("restart"),
    "restart_advanced": _get_icon("restart_advanced"),
    "robot_error": _get_icon("robot_error"),
    "search": _get_icon("search"),
    "settings": _get_icon("settings"),
    "shell": _get_icon("shell"),
    "shutdown": _get_icon("shutdown"),
    "sleep": _get_icon("sleep"),
    "sourcecode": _get_icon("sourcecode"),
    "stackoverflow": _get_icon("stackoverflow"),
    "twitter": _get_icon("twitter"),
    "up": _get_icon("up"),
    "update": _get_icon("update"),
    "url": _get_icon("url"),
    "user": _get_icon("user"),
    "warning": _get_icon("warning"),
    "web_search": _get_icon("web_search"),
    "wiki": _get_icon("wiki"),
    "windowsindexingoptions": _get_icon("windowsindexingoptions"),
    "windowssettingslight": _get_icon("windowssettings.light"),
    "wizard": _get_icon("wizard"),
    "wolframalpha": _get_icon("wolframalpha"),
    "work": _get_icon("work"),
    "yahoo": _get_icon("yahoo"),
    "youtube": _get_icon("youtube"),
    "youtubemusic": _get_icon("youtubemusic"),
}



import logging
import sys
from pathlib import Path
from typing import Any

_logger = logging.getLogger(__name__)


def logger(obj: Any) -> logging.Logger:
    module_file = sys.modules[obj.__module__].__file__
    if module_file is not None:
        module_name = Path(module_file).stem
        return logging.getLogger(f"{module_name}.{obj.__class__.__name__}")
    return logging.getLogger(f"{obj.__module__}.{obj.__name__}")



from typing import TypedDict, Literal

MANIFEST_FILE = 'plugin.json'

Languages = Literal[
    'Python',
    'CSharp',
    'FSharp',
    'Executable',
    'TypeScript',
    'JavaScript',
]


class PluginManifestSchema(TypedDict):
    ID: str
    ActionKeyword: str
    Name: str
    Description: str
    Author: str
    Version: str
    Language: Languages
    Website: str
    IcoPath: str
    ExecuteFileName: str



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

from .result import JsonRPCAction, Result, ResultResponse, send_results
from .shared import logger


class Method(ABC):

    def __init__(self) -> None:
        self._logger = logger(self)
        self._results: list[Result] = []

    def add_result(self, result: Result) -> None:
        self._results.append(result)

    def return_results(self, settings: Optional[Dict[str, Any]] = None) -> ResultResponse:
        return send_results(self._results, settings)

    @abstractmethod
    def __call__(self, *args, **kwargs) -> ResultResponse | JsonRPCAction:
        pass



from typing import Optional

from .result import JsonRPCAction

NAME_SPACE = 'Flow.Launcher'


def _send_action(method: str, *parameters) -> JsonRPCAction:
    return {"method": f"{NAME_SPACE}.{method}", "parameters": parameters}


def change_query(query: str, requery: bool = False) -> JsonRPCAction:
    """Change the query in Flow Launcher."""
    return _send_action("ChangeQuery", query, requery)


def shell_run(command: str, filename: str = 'cmd.exe') -> JsonRPCAction:
    """Run a shell command."""
    return _send_action("ShellRun", command, filename)


def close_app() -> JsonRPCAction:
    """Close Flow Launcher."""
    return _send_action("CloseApp")


def hide_app() -> JsonRPCAction:
    """Hide Flow Launcher."""
    return _send_action("HideApp")


def show_app() -> JsonRPCAction:
    """Show Flow Launcher."""
    return _send_action("ShowApp")


def show_msg(title: str, sub_title: str, ico_path: str = "") -> JsonRPCAction:
    """Show a message in Flow Launcher."""
    return _send_action("ShowMsg", title, sub_title, ico_path)


def open_setting_dialog() -> JsonRPCAction:
    """Open the settings window in Flow Launcher."""
    return _send_action("OpenSettingDialog")


def start_loading_bar() -> JsonRPCAction:
    """Start the loading bar in Flow Launcher."""
    return _send_action("StartLoadingBar")


def stop_loading_bar() -> JsonRPCAction:
    """Stop the loading bar in Flow Launcher."""
    return _send_action("StopLoadingBar")


def reload_plugins() -> JsonRPCAction:
    """Reload the plugins in Flow Launcher."""
    return _send_action("ReloadPlugins")


def copy_to_clipboard(text: str, direct_copy: bool = False, show_default_notification=True) -> JsonRPCAction:
    """Copy text to the clipboard."""
    return _send_action("CopyToClipboard", text, direct_copy, show_default_notification)


def open_directory(directory_path: str, filename_or_filepath: Optional[str] = None) -> JsonRPCAction:
    """Open a directory."""
    return _send_action("OpenDirectory", directory_path, filename_or_filepath)


def open_url(url: str, in_private: bool = False) -> JsonRPCAction:
    """Open a URL."""
    return _send_action("OpenUrl", url, in_private)


def open_uri(uri: str) -> JsonRPCAction:
    """Open a URI."""
    return _send_action("OpenAppUri", uri)
